/***************************************************************************
                     uncouple_channels.c  -  description
                             -------------------
   Restore the high frequency portion of the coupled channels.
                             -------------------
    begin                : Fri Oct 14 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "uncouple_channels.h"


void uncouple_channels(AUDIO_BLOCK *blk)
{
   u8        *cpl_bapexp, *cpl_coord;
   CH_STRUCT *ch;
   AC3_COEFF *src_coeff, *dst_coeff;
   u32        cnt, num_bins, bin_start, cpl_sbnds;
   AC3_COEFF  cpl_coeff;
   u8         dither, gain, cpl_exp, chincpl;

   /* copy mantissas to all coupled channels */
   num_bins     = (u32)blk->cpl_data.num_bins;
   bin_start    = (u32)blk->cpl_data.bin_start;
   cpl_bapexp   = blk->cpl_data.bap;
   ch           = blk->ch_data + blk->cpl_data.ch;
   dither       = blk->dither << blk->cpl_data.ch;
   src_coeff    = ch->coeff + bin_start;
   while (*(ch->cpl_coord) < 0x80) {
      ch++;
      if (*(ch->cpl_coord)) {
         dst_coeff = ch->coeff + bin_start;
         cnt       = num_bins;
         do {
            if (*(cpl_bapexp++) == 0) {
               if (dither >= 0x80)
                  GET_DITHER_VALUE(*dst_coeff, blk->rng)
               else
                  *dst_coeff = 0;
            }
            else
               *dst_coeff = *src_coeff;
            src_coeff++;
            dst_coeff++;
         } while (--cnt);
         cpl_bapexp -= num_bins;
         src_coeff  -= num_bins;
      }
      dither <<= 1;
   }

   /* restore high frequency portions of each channel */
   cpl_bapexp -= AC3_MAX_CPL_BIN+CPL_PAD_BYTES;
   cpl_coord   = (u8 *)(src_coeff - bin_start);
   cpl_coord  -= 2+AC3_NUM_BND;
   do {
#define INVERT    bin_start
#define RSHIFT    dither
      cpl_sbnds = blk->coupled_sbands;
      INVERT    = blk->phase_invert;
      chincpl   = *(cpl_coord++);
      gain      = *(cpl_coord++) * 3;
#if (AC3_COEFF_SIZE > 29)
      gain     += 5;
#endif
      cnt       = num_bins;
      dst_coeff = src_coeff;
      goto GET_NEW_CPL_COEFF; /* skip unecessary tests */
      do {
         if ((cnt % AC3_BIN_PER_SBND) == 0) {
            if (cpl_sbnds < 0x80000000) {
GET_NEW_CPL_COEFF:
               cpl_exp    = *cpl_coord >> 4;
               cpl_coeff  = (AC3_COEFF)(*(cpl_coord++) & 0x0F);
#if (AC3_COEFF_SIZE > 29)
               if (cpl_exp < 15) {
                  cpl_exp++;
                  cpl_coeff += 16;
               }
               cpl_coeff <<= AC3_COEFF_SIZE - cpl_exp - gain;
#else
               if (cpl_exp == 15)
                  cpl_coeff <<= AC3_COEFF_SIZE - 5;
               else {
                  cpl_coeff  += 16;
                  cpl_coeff <<= AC3_COEFF_SIZE - 6;
               }
               cpl_coeff >>= cpl_exp + gain;
#endif
               if (INVERT >= 0x80000000)
                  cpl_coeff = ~cpl_coeff;    /* phase inversion */
               INVERT <<= 1;
            }
            cpl_sbnds <<= 1;
         }

         if ( (!cpl_coeff)||(!(*dst_coeff)) )
            *dst_coeff = 0;
         else {
            cpl_exp = (AC3_COEFF_SHIFT - 3) + *cpl_bapexp;
            if (cpl_exp >= (AC3_COEFF_SIZE-1)) {
               *dst_coeff  ^= cpl_coeff;
               *dst_coeff >>= AC3_COEFF_SIZE-1;
            }
            else {
               *dst_coeff = (AC3_COEFF)AC3_COEFF_MULT_SHIFT(*dst_coeff,
                                                             cpl_coeff,
                                                             cpl_exp);
            }
         }
         dst_coeff++;
         cpl_bapexp++;
      } while (--cnt);
      cpl_bapexp -= num_bins;
      cpl_coord  += sizeof(CH_STRUCT) - 2 - (u32)blk->num_coupled_bands;
      src_coeff  += sizeof(CH_STRUCT) / sizeof(AC3_COEFF);
#undef INVERT
   } while (chincpl < 0x80);
}

