/***************************************************************************
                          AVindex.h  -  description
                             -------------------
   Functions to store and access the index of current AV stream.
                             -------------------
    begin                : Thu Jan 26 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __AVINDEX_H__
#define __AVINDEX_H__

#include "common/File.h"
#include "MPEG2_SYS.h"

#define AVINDEX_INTERVAL   (18*1024*1024)

#ifdef __cplusplus
extern "C" {
#endif

/* Returns the number of bytes of memory needed by the AVindex. */
size_t AVindex_mem_needed();

/* Uses the memory pointed to by mem, for the AVindex. */
void AVindex_set_mem(void *mem);

/* Clear the index, and put the given data at the begining of
    the index.
      INPUT: f_pos - File position of the first entry.
             v_pos - Time position of the video stream.
             a_pos - Time position of the audio stream.
             v_bck - Bytes behind f_pos, where v_pos occurred.
             a_bck - Bytes behind f_pos, where a_pos occurred. */
void AVindex_init(FILE_POS f_pos,
                  CLK_VAL  v_pos, CLK_VAL a_pos,
                  u32      v_bck, u32     a_bck);

/* Add a new entry to the index.
      INPUT: f_pos - File position of the new entry.
             v_pos - Time position of the video stream.
             a_pos - Time position of the audio stream.
             v_bck - Bytes behind f_pos, where v_pos occurred.
             a_bck - Bytes behind f_pos, where a_pos occurred. */
void AVindex_add(FILE_POS f_pos,
                 CLK_VAL  v_pos, CLK_VAL a_pos,
                 u32      v_bck, u32     a_bck);

/* Get an entry from the index.
      INPUT:  v_pos - Video position to search for. NULL if not needed.
              a_pos - Audio position to search for. NULL if not needed.
      OUTPUT: f_pos - File position of the requested entry.
              v_pos - Time position of the video stream.
              a_pos - Time position of the audio stream.
              v_bck - Bytes behind f_pos, where v_pos occurred.
              a_bck - Bytes behind f_pos, where a_pos occurred.
      RETURN: Success   - non-zero value
              Not found - 0 */
int AVindex_get(FILE_POS *f_pos,
                CLK_VAL  *v_pos, CLK_VAL *a_pos,
                u32      *v_bck, u32     *a_bck);

#if 0
def __DVD_INPUT__
/* Get an entry from the index.
      INPUT:  f_pos - File position to search for.
      OUTPUT: f_pos - File position of the requested entry.
              v_pos - Time position of the video stream.
              a_pos - Time position of the audio stream.
              v_bck - Bytes behind f_pos, where v_pos occurred.
              a_bck - Bytes behind f_pos, where a_pos occurred.
      RETURN: Success   - non-zero value
              Not found - 0 */
int AVindex_find_pos(FILE_POS *f_pos,
                     CLK_VAL  *v_pos, CLK_VAL *a_pos,
                     u32      *v_bck, u32     *a_bck);
#endif

#ifdef __cplusplus
};
#endif

#endif

