/***************************************************************************
                          PES_PKT.h  -  description
                             -------------------
   Stuff needed for handling PES packets.
                             -------------------
    begin                : Sat Sep 4 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PES_PKT_H__
#define __PES_PKT_H__

#include "common/typedefs.h"
#include "MPEG2_SYS.h"


#define PES_MAX_HEADER_SIZE  264

#define PES_PROGRAM_STREAM_MAP        0xBC
#define PES_PRIVATE_STREAM_1          0xBD
#define PES_ATSC_AC3_AUDIO            0xBD
#define PES_PADDING_STREAM            0xBE
#define PES_PRIVATE_STREAM_2          0xBF
#define PES_ECM_STREAM                0xF0
#define PES_EMM_STREAM                0xF1
#define PES_ISO_13818_1A_STREAM       0xF2
#define PES_ISO_13818_6_DSMCC_STREAM  0xF2
#define PES_ISO13522_STREAM           0xF3
#define PES_ITUT_H2221_A              0xF4
#define PES_ITUT_H2221_B              0xF5
#define PES_ITUT_H2221_C              0xF6
#define PES_ITUT_H2221_D              0xF7
#define PES_ITUT_H2221_E              0xF8
#define PES_ANCILLARY_STREAM          0xF9
#define PES_PROGRAM_STREAM_DIRECTORY  0xFF

#define PES_MPEG_AUDIO_STREAM(ptr_to_PES_PKT) \
   ( (0xBF < ptr_to_PES_PKT->stream_id)&&(ptr_to_PES_PKT->stream_id < 0xE0) )
#define PES_MPEG_VIDEO_STREAM(ptr_to_PES_PKT) \
   ( (0xDF < ptr_to_PES_PKT->stream_id)&&(ptr_to_PES_PKT->stream_id < 0xF0) )
#define PES_RESERVED_STREAM(ptr_to_PES_PKT) \
   ( (0xF9 < ptr_to_PES_PKT->stream_id)&&(ptr_to_PES_PKT->stream_id < 0xFF) )

#define PES_MPEG_AV_STREAM_NUM(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->stream_id & 0x0F)
#define PES_SCRAMBLING_CONTROL(ptr_to_PES_PKT) \
   ((ptr_to_PES_PKT->flags1 >> 4) & 0x03)
#define PES_PRIORITY_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags1 & 0x08)
#define PES_DATA_ALIGN_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags1 & 0x04)
#define PES_COPYRIGHT_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags1 & 0x02)
#define PES_ORIGINAL_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags1 & 0x01)
#define PES_PTS_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags2 & 0x80)
#define PES_DTS_FLAG(ptr_to_PES_PKT) \
   ((ptr_to_PES_PKT->flags2 & 0xC0) == 0xC0)
#define PES_ESCR_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags2 & 0x20)
#define PES_ES_RATE_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags2 & 0x10)
#define PES_TRICK_MODE_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags2 & 0x08)
#define PES_MORE_COPY_INFO_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags2 & 0x04)
#define PES_CRC_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags2 & 0x02)
#define PES_EXTENSION_FLAG(ptr_to_PES_PKT) \
   (ptr_to_PES_PKT->flags2 & 0x01)
   

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8  stream_id,               /*  8: Identifies the type and number of this
                                         elementary stream. */
       flags1,
/*
      2   <'10'>
      2   PES_scrambling_control       Indicates the scrambling mode of this
                                         packet's payload. See
                                         transport_scrambling_control for
                                         values.
                                  <'00'> Always use this value in ATSC streams.
      1   PES_priority                 Set when the payload of this packet has
                                         a higher priority than other PES
                                         packets a PES_priority value of 0.
      1   data_align_indicator         Set when the payload of this packet
                                         starts with the video start code, or
                                         audio sync word. When not set, the
                                         payload may, or may not be aligned.
                               <1> Always use this value in ATSC video streams.
      1   copyright                    Set if the material in this stream is
                                         protected by coptright.
      1   original_or_copy             Set if the material in this stream is
                                         original. When not set, the material
                                         was copied.
*/
       flags2,
/*
      2   PTS_DTS_flags  <'00'> No PTS or DTS fields are in this packet
                         <'01'> forbidden
                         <'10'> The PTS field is in this packet
                         <'11'> PTS and DTS fields are in this packet
      1   ESCR_flag                    Set when the ESCR section is in this
                                         packet.
                     <0> Always use this value in ATSC streams.
      1   ES_rate_flag                 Set when the ES_rate field is in this
                                         packet.
                        <0> Always use this value in ATSC streams.
      1   DSM_trick_mode_flag          Set when the trick_mode_control field
                                         is in this packet.
      1   additional_copy_info_flag    Set when the additional_copy_info
                                         field is in this packet.
      1   PES_CRC_flag                 Set when the prev_PES_packet_CRC
                                         field is in this packet.
                        <0> Always use this value in ATSC streams.
      1   PES_extension_flag           Set when the extension field is in
                                         this packet.
*/
       header_length;           /*  8: Number of bytes, in this PES packet
                                         header, that follow this one. */

   u16 packet_length,           /* 16: Number of bytes following this HEADER.
                                         A value of 0 indicates this
                                         PES_PACKET has an unspecified, or
                                         unbounded length. The 0 length can
                                         only be used in video streams. */
/*
 * ATSC video streams always have packet_length set to 0. Each PES_PKT
 * contains only 1 video access unit. For terrestrial brodcast, no more than
 * 1 coded video frame is in each packet. When there is a  continuity_counter
 * discontinuity (discontinuity_indicator == 1), no frames will be present in
 * the packet.
 */
       payload_len16,           /*     Number of payload bytes. */

       stream_id2,
       state;

   CLK_VAL PTS;                 /*     Presentation Time Stamp. The time at
                                         which the audio or video, in this
                                         packet must be presented to the
                                         audience. For audio streams, the
                                         PTS refers to the first access unit
                                         in this packet. For video streams,
                                         the PTS refers to the access unit
                                         containing the first byte of the
                                         picture start code.
          For audio access units, and video access units with the
          low_delay_flag set, or containing B-frames:
            PTS = DTS (decode and present immediately)
          For video access units containing I or P frames, and
          low_delay_flag cleared:
            PTS = DTS of next I or P frame (decode and delay presentation)
   ATSC audio and video streams always have a PTS value. The DTS values is
   included when PTS is not equal to DTS. */
/*       DTS;                            Decoding Time Stamp. The time at
                                         which the audio or video, in this
                                         packet must be decoded. For video
                                         streams, the DTS refers to the
                                         access unit containing the first
                                         byte of the picture start code. The
                                         DTS value is only needed when DTS is
                                         not equal to PTS. */

   const
   u8 *payload;
} PES_PKT;


/* Reads the first PES packet header, in data_stream, into pes.
      INPUT: data_stream - Points to the data stream holding a PES packet.
             stream_len  - Number of bytes in the data stream.
      OUTPUT: pes - Points to the PES_PKT that will hold data taken from
                      data_stream.
      RETURN: Finished reading packet - Non-zero value.
              More data is needed     - 0 */
u32 PES_PKT_read(const u8 *data_stream, 
                 u32       stream_len, 
                 PES_PKT  *pes);


#ifdef __cplusplus
};
#endif

#endif
