/***************************************************************************
                          PIC_HEADER.h  -  description
                             -------------------
   Stuff needed for handling MPEG1 picture and gop headers.
                             -------------------
    begin                : Tue Sep 28 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PIC_HEADER1_H__
#define __PIC_HEADER1_H__

#include "common/typedefs.h"


#define GOP_HEADER1_DROP_FRAME_FLAG(ptr_to_PIC_HEADER1) \
   (ptr_to_PIC_HEADER1->type & 0x04)
#define GOP_HEADER1_CLOSED_GOP_FLAG(ptr_to_PIC_HEADER1) \
   (ptr_to_PIC_HEADER1->type & 0x02)
#define GOP_HEADER1_BROKEN_LINK_FLAG(ptr_to_PIC_HEADER1) \
   (ptr_to_PIC_HEADER1->type & 0x01)

#define PIC_HEADER1_CODING_TYPE(ptr_to_PIC_HEADER1) \
   (ptr_to_PIC_HEADER1->type & 0x70)
#define CODING_TYPE_I_FRAME   0x10
#define CODING_TYPE_P_FRAME   0x20
#define CODING_TYPE_B_FRAME   0x30
#define CODING_TYPE_D_FRAME   0x40
#define CODING_TYPE_FORBIDDEN(ptr_to_PIC_HEADER1) \
   (!(ptr_to_PIC_HEADER1->type & 0x70))
#define CODING_TYPE_RESERVED(ptr_to_PIC_HEADER) \
   ((ptr_to_PIC_HEADER->type & 0x70) > 0x40)

#define FULL_PIXEL_FLAG    0x10


#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8 type,                 /* <'000'> forbidden
                               <'001'> I_FRAME (intra-coded)
                               <'010'> P_FRAME (predictive-coded)
                               <'011'> B_FRAME (bidirectionally-predictive-coded)
                               <'100'> D_FRAME (these can be ignored)
                         <'101'-'111'> reserved */
/* GOP_HEADER FLAGS
      1   drop_frame_flag              Set when the frame rate is 29.97fps.
                                         When not set the frame rate is
                                         rounded to the nearest integer.
      1   closed_gop                   Set when the first set of B-frames,
                                         after the I-frame following this
                                         GOP_HEADER, have been encoded using
                                         only backward prediction (the
                                         B-frames can be decoded, if you start
                                         decoding from the following I-frame).
                                         When not set, the decoder must skip
                                         the B-frames, if it starts decoding
                                         from the following I-frame.
      1   broken_link                  Set when the first set of B-frames,
                                         after the I-frame following this
                                         GOP_HEADER, cannot be decoded
                                         because the forward prediction frame
                                         was removed from the sequence. After
                                         initial encoding, this bit is 0.
                                         The value can change during the
                                         editing process.
*/
       unused,
       f_code[2];               /*     bit 4, of the f_code bytes, is used
                                         for the full pixel flag. */
/*
      3   f_code[0]                    forward
      3   f_code[1]                    backward
 *
 *        f_code[s]             <0x00> forbidden
 *        (unsigned byte)  <0x01-0x07> Valid values for P_FRAMEs(forward) and B_FRAMEs.
 *                         <0x08-0xFF> forbidden (these are limited to 3-bit values!)
 */
   u16 temporal_reference,      /* 10: Indicates the order in which this
                                         picture will be displayed relative
                                         to the others in the sequence.*/
       vbv_delay;               /* 16: Delay between the time the last byte of
                                         the PIC_START_CODE is received, and
                                         the time this picture begins to be
                                         decoded. Measured in periods of the
                                         90kHz (27MHz/300) system derived
                                         clock. If one picture, in the
                                         sequence, has a vbv_delay value of
                                         0xFFFF, then all pictures must have
                                         the same value. */
} PIC_HEADER1;


/* Reads the gop header in data_stream.
      INPUT: data_stream - Points to the data stream holding the gop
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER1 that will hold the data.
              cur_pos - Used by MPEG1_VIDEO_MODULE.
              code    - Used by MPEG1_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 GOP_HEADER1_read(const u8          *data_stream,
                           u32          data_len,
                           PIC_HEADER1 *hdr,
                           u32         *cur_pos,
                           u16         *code);

/* Reads the picture header in data_stream.
      INPUT: data_stream - Points to the data stream holding the picture
                             header. This stream does not have to
                             contain the entire header. Repeated calls to
                             this function can be used to read the header
                             in parts.
             data_len    - Number of bytes in the data stream.
      OUTPUT: hdr     - Points to the PIC_HEADER1 that will hold the data.
              cur_pos - Used by MPEG1_VIDEO_MODULE.
              code    - Used by MPEG1_VIDEO_MODULE.
      RETURN: Number of bytes read. */
u32 PIC_HEADER1_read(const u8          *data_stream,
                           u32          data_len,
                           PIC_HEADER1 *hdr,
                           u32         *cur_pos,
                           u16         *code);
#ifdef __cplusplus
};
#endif

#endif
