/***************************************************************************
                           dct32.h  -  description
                             -------------------
    begin                : Tue Apr 26 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _DCT32_H_
#define _DCT32_H_

#include "common/typedefs.h"

/* For slow, full precision audio use: #define FP_SHIFT  23
   For fast, less precise audio use:   #define FP_SHIFT  12 */
#define FP_SHIFT           12
#define FP_CONST(fp_value) (((fp_value) * (1 << FP_SHIFT)) + 0.5)

#if (FP_SHIFT <= 12)
typedef s16 AV1;
typedef s32 AV2;
#else
typedef s32 AV1;
typedef s64 AV2;
#endif

#if defined(__MMX__) && (FP_SHIFT <= 12)
# if defined(__GCC__)
#  define DCT32_DONE    __asm__ __volatile__("emms")
# elif defined(__MSVC__)
#  define DCT32_DONE    __asm emms
# else
#  error "Don't know how to insert inline ASM code for this compiler"
# endif
#else
#  define DCT32_DONE
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Does a 32pt DCT operation.
      INPUT: data - An array holding the 32 sub-band samples.
      OUTPUT: data - The 32 samples are replaced by 32 dct
                       coefficients. */
void dct32(AV1 *data);

#ifdef __cplusplus
};
#endif

#endif

