/***************************************************************************
                         inv_trans.h  -  description
                             -------------------
    begin                : Fri Jul 28 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define VC1_VIDEO_TRACE*/
#ifdef VC1_VIDEO_TRACE
# include <stdio.h>
# include "PIC_LAYER.h"

/*# define OVERLAP_TRACE*/
#endif

#ifdef OVERLAP_TRACE
# define OVERLAP_TRACE_INPUT(v0, v1, v2, v3, c0, c1) \
   VC1_TRACE_BEG_BLK \
      printf("%hd %hd %hd %hd  (%hd %hd)\n", v0, v1, v2, v3, c0, c1); \
   VC1_TRACE_END_BLK

# define OVERLAP_TRACE_OUTPUT(v0, v1, v2, v3) \
   VC1_TRACE_BEG_BLK \
      printf("%hd %hd %hd %hd\n", v0, v1, v2, v3); \
   VC1_TRACE_END_BLK
#else
# define OVERLAP_TRACE_INPUT(v0, v1, v2, v3, c0, c1)
# define OVERLAP_TRACE_OUTPUT(v0, v1, v2, v3)
#endif

#include "inv_trans.h"

const VC1_INV_TRANS_FUNC VC1_inv_trans[] = {VC1_8x8_inv_trans,
                                            VC1_8x4_inv_trans,
                                            VC1_4x8_inv_trans,
                                            VC1_4x4_inv_trans};

static const s16 table_8x8[] = {16,  15,   9,   4,
                                15,  -4, -16,  -9,
                                 9, -16,   4,  15,
                                 4,  -9,  15, -16};


#define TRANS_ROW_OP_8x1(inc, C, sh) \
j = 8; \
do {                 /* data[0]  out[0]  tab[0] */ \
   i     = 8; \
   val   = C + (12 * (*data)); \
   data += inc; \
   do { \
      *(out++) = val; \
   } while (--i);    /* data[1]  out[8]  tab[0] */ \
   TRANS_OP_8x1(inc, sh)

#define TRANS_COL_OP_8x1(inc, C, sh) \
j = 8; \
do {                 /* data[0]  out[0]  tab[0] */ \
   i     = 4; \
   val   = C + (12 * (*data)); \
   data += inc; \
   do { \
         *(out++) = val; \
   } while (--i); \
   i = 4; \
   val++; \
   do { \
         *(out++) = val; \
   } while (--i);    /* data[1]  out[8]  tab[0] */ \
   TRANS_OP_8x1(inc, sh)

#define TRANS_OP_8x1(inc, sh) \
   i = 7; \
   do { \
      val       = (*(tab++)) * (*data); \
      *(--out) -= val; \
      *(out-i) += val; \
      i        -= 2; \
   } while (i > 0);  /* data[1]  out[4]  tab[4] */ \
   data     += inc; \
   val       = (*data) << 4; \
                     /* data[2]  out[4]  tab[4] */ \
   *(out--) -= val; \
   *(out--) -= val; \
   *(out-2) += val; \
   *(out+5) += val; \
   val       = (*data) * 6; \
   data     += inc; \
   *(out--) -= val; \
   *(out--) += val; \
   *(out+5) -= val; \
   *(out+6) += val; \
                     /* data[3]  out[0]  tab[4] */ \
   i = 6; \
   do { \
      val       = (*(tab++)) * (*data); \
      *(out++) += val; \
      *(out+i) -= val; \
      i        -= 2; \
   } while (i >= 0); /* data[3]  out[4]  tab[8] */ \
   out  -= 4; \
   data += inc; \
   val   = 12 * (*data); \
                     /* data[4]  out[0]  tab[8] */ \
   *(out++) += val; \
   *(out++) -= val; \
   *(out++) -= val; \
   *(out++) += val; \
   *(out++) += val; \
   *(out++) -= val; \
   *(out++) -= val; \
   *(out++) += val; \
                     /* data[4]  out[8]  tab[8] */ \
   i     = 7; \
   data += inc; \
   do {              /* data[5]  out[8]  tab[8] */ \
      val       = (*(tab++)) * (*data); \
      *(--out) -= val; \
      *(out-i) += val; \
      i        -= 2; \
   } while (i > 0);  /* data[5]  out[4]  tab[12] */ \
   data     += inc; \
   val       = (*data) * 6; \
                     /* data[6]  out[4]  tab[12] */ \
   *(out--) -= val; \
   *(out--) -= val; \
   *(out-2) += val; \
   *(out+5) += val; \
   val       = (*data) << 4; \
   data     += inc; \
   *(out--) += val; \
   *(out--) -= val; \
   *(out+5) += val; \
   *(out+6) -= val; \
                     /* data[7]  out[0]  tab[12] */ \
   i = 6; \
   do { \
      val       = (*(tab++)) * (*data); \
      *(out++) += val; \
      *(out+i) -= val; \
      i        -= 2; \
   } while (i >= 0); /* data[7]  out[4]  tab[16] */ \
   out +=  4; \
   tab -= 16; \
   i    =  8; \
   do { \
      *data = (*(--out)) >> sh; \
      data -= inc; \
   } while (--i); \
   data += 9; \
} while (--j)


#define TRANS_OP_4x2(inc, C, sh) \
j = 8; \
do {                 /* data[0]  out[0] */ \
   i     = 4; \
   val   = C + (17 * (*data)); \
   data += inc*4; \
   do { \
      *(out++) = val; \
   } while (--i);    /* data[4]  out[4] */ \
   i     = 4; \
   val   = C + (17 * (*data)); \
   data += inc; \
   do { \
      *(out++) = val; \
   } while (--i);    /* data[5]  out[8] */ \
   val       = (*data) * 22; \
   *(--out) -= val; \
   *(out-3) += val; \
   val       = (*data) * 10; \
   data     -= inc*4; \
   *(--out) -= val; \
   *(--out) += val; \
   out--;            /* data[1]  out[4] */ \
   val       = (*data) * 22; \
   *(--out) -= val; \
   *(out-3) += val; \
   val       = (*data) * 10; \
   data     += inc; \
   *(--out) -= val; \
   *(--out) += val; \
   out--;            /* data[2]  out[0] */ \
   val       = (*data) * 17; \
   data     += inc*4; \
   *(out++) += val; \
   *(out++) -= val; \
   *(out++) -= val; \
   *(out++) += val;  /* data[6]  out[4] */ \
   val       = (*data) * 17; \
   data     += inc; \
   *(out++) += val; \
   *(out++) -= val; \
   *(out++) -= val; \
   *(out++) += val;  /* data[7]  out[8] */ \
   val       = (*data) * 10; \
   *(--out) -= val; \
   *(out-3) += val; \
   val       = (*data) * 22; \
   data     -= inc*4; \
   *(--out) += val; \
   *(--out) -= val; \
   out--;            /* data[3]  out[4] */ \
   val       = (*data) * 10; \
   *(--out) -= val; \
   *(out-3) += val; \
   val       = (*data) * 22; \
   *(--out) += val; \
   *(--out) -= val;  /* data[3]  out[1] */ \
   out  += 7; \
   data += inc*4; \
   i     = 8; \
   do { \
      *data = (*(--out)) >> sh; \
      data -= inc; \
   } while (--i); \
   data += 9; \
} while (--j)


/* Does the VC1 8x8 inverse transformation.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_8x8_inv_trans(s16 *data)
{
   s32  tmp[8];
   s32 *out;
   const
   s16 *tab;
   s32  j, i, val;

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("VC1_8x8_inv_trans()\n");
   VC1_TRACE_END_BLK
#endif

   /* process rows */
   out = tmp;
   tab = table_8x8;
   TRANS_ROW_OP_8x1(1, 4, 3);

   /* process cols */
   data -= 64;
   TRANS_COL_OP_8x1(8, 64, 7);
}

/* Does 2 VC1 8x4 inverse transformations.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_8x4_inv_trans(s16 *data)
{
   s32  tmp[8];
   s32 *out;
   const
   s16 *tab;
   s32  j, i, val;

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("VC1_8x4_inv_trans()\n");
   VC1_TRACE_END_BLK
#endif

   /* process rows */
   out = tmp;
   tab = table_8x8;
   TRANS_ROW_OP_8x1(1, 4, 3);

   /* process cols */
   data -= 64;
   TRANS_OP_4x2(8, 64, 7);
}

/* Does 2 VC1 4x8 inverse transformations.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_4x8_inv_trans(s16 *data)
{
   s32  tmp[8];
   s32 *out;
   const
   s16 *tab;
   s32  j, i, val;

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("VC1_4x8_inv_trans()\n");
   VC1_TRACE_END_BLK
#endif

   /* process rows */
   out = tmp;
   TRANS_OP_4x2(1, 4, 3);

   /* process cols */
   tab   = table_8x8;
   data -= 64;
   TRANS_COL_OP_8x1(8, 64, 7);
}

/* Does 4 VC1 4x4 inverse transformation.
      INPUT: data - An array holding the 64 coefficients.
      OUTPUT: data - The 64 coefficients are replaced by 64
                       recovered samples. */
void VC1_4x4_inv_trans(s16 *data)
{
   s32  tmp[8];
   s32 *out;
   s32  j, i, val;

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("VC1_4x4_inv_trans()\n");
   VC1_TRACE_END_BLK
#endif

   /* process rows */
   out = tmp;
   TRANS_OP_4x2(1, 4, 3);

   /* process cols */
   data -= 64;
   TRANS_OP_4x2(8, 64, 7);
}


#ifdef USE_PRECISE_OVERLAP

# define OVERLAP_OP2(t0, t1, c0, c1, i, ii) \
OVERLAP_TRACE_INPUT(pre[0], pre[1], blk[0], blk[i], c0, c1) \
res[0] = c0+1024;  res[1] = c1+1024;  res[2] = c0;       res[3] = c1; \
res[0]+= 7*pre[0]; res[1]-=   pre[0]; res[2]+=   pre[0]; res[3]+=   pre[0]; \
                   res[1]+= 7*pre[1]; res[2]+=   pre[1]; \
                   res[1]+=   blk[0]; res[2]+= 7*blk[0]; \
res[0]+=   blk[i]; res[1]+=   blk[i]; res[2]-=   blk[i]; res[3]+= 7*blk[i]; \
res[0]>>= 3;       res[1]>>= 3;       res[2]>>= 3;       res[3]>>= 3; \
OVERLAP_TRACE_OUTPUT((res[0]-128), (res[1]-128), res[2], res[3]) \
blk[0] = res[2]; \
blk[i] = res[3]; \
pre   +=  2; \
blk   += ii; \
if (res[0] <= 0) \
   t0 = 0; \
else if (res[0] >= 255) \
   t0 = 255; \
else \
   t0 = (u8)res[0]; \
if (res[1] <= 0) \
   t1 = 0; \
else if (res[1] >= 255) \
   t1 = 255; \
else \
   t1 = (u8)res[1]


/* Does the overlap filtering operation in the vertical direction. */
void VC1_overlap_Y(s16 *blk, s16 *pre, s16 *preB, s16 *preT, u8 *outp, int inc)
{
   s16 res[4];
   u16 rows, i;

   if (outp) {
      if (preB) {
         i = 2;
         preB -= 4;
      }
      else {
         i = 4;
         preB = preT + 32+32-8;
      }
      preT += 32-6;
# ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("Poverlap_Y()\n");
      VC1_TRACE_END_BLK
# endif
      outp -= 2;
      rows  = 4;
      do {
         /* even rows */
         OVERLAP_OP2(outp[0], outp[1], 4, 3, 1, 8);
         outp += inc;
         if (rows == 1) {
            preB[0] = res[0]-128;
            preB[i] = res[1]-128;
         }
         else if (rows == 4) {
            preT[0] = res[0]-128;
            preT[4] = res[1]-128;
         }

         /* odd rows */
         OVERLAP_OP2(outp[0], outp[1], 3, 4, 1, 8);
         outp += inc;
         if (rows == 1) {
            preB[1]   = res[0]-128;
            preB[i+1] = res[1]-128;
         }
         else if (rows == 4) {
            preT[1] = res[0]-128;
            preT[5] = res[1]-128;
         }
      } while (--rows);
      blk -= 64-6;
      pre -= 16;
   }
   else
      blk += 6;

   /* store right edge */
   rows = 8;
   do {
      *(pre++) = blk[0];
      *(pre++) = blk[1];
      blk += 8;
   } while (--rows);
}

/* Does the overlap filtering operation in the horizontal direction. */
void VC1_overlap_X(s16 *blk, s16 *pre, s16 *pre2, u8 *outp, int inc)
{
   u8  *out2;
   s16  res[4];
   u16  cols;

   if (outp) {
# ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("Poverlap_X()\n");
      VC1_TRACE_END_BLK
# endif
      out2 = outp - inc;
      outp = out2 - inc;
      cols = 4;
      do {
         /* even cols */
         OVERLAP_OP2(outp[0], out2[0], 4, 3, 8, 1);
         outp++;
         out2++;

         /* odd cols */
         OVERLAP_OP2(outp[0], out2[0], 3, 4, 8, 1);
         outp++;
         out2++;
      } while (--cols);
      blk -=  8-48;
      pre -= 16;
   }
   else
      blk += 48;

   /* store bottom edge */
   cols = 8;
   do {
      *(pre2++) = *pre;
      *(pre++)  = blk[0];
      *(pre2++) = *pre;
      *(pre++)  = blk[8];
      blk++;
      pre2 += 2;
   } while (--cols);
}


# define OVERLAP_OP4(t0, t1, t2, t3, c0, c1) \
OVERLAP_TRACE_INPUT(pre[0], pre[1], pre[2], pre[3], c0, c1) \
res[0] = c0+1024;  res[1] = c1+1024;  res[2] = c0+1024;  res[3] = c1+1024; \
res[0]+= 7*pre[0]; res[1]-=   pre[0]; res[2]+=   pre[0]; res[3]+=   pre[0]; \
                   res[1]+= 7*pre[1]; res[2]+=   pre[1]; \
                   res[1]+=   pre[2]; res[2]+= 7*pre[2]; \
res[0]+=   pre[3]; res[1]+=   pre[3]; res[2]-=   pre[3]; res[3]+= 7*pre[3]; \
res[0]>>= 3;       res[1]>>= 3;       res[2]>>= 3;       res[3]>>= 3; \
OVERLAP_TRACE_OUTPUT((res[0]-128), (res[1]-128), (res[2]-128), (res[3]-128)) \
pre += 4; \
if (res[0] <= 0) \
   *t0 = 0; \
else if (res[0] >= 255) \
   *t0 = 255; \
else \
   *t0 = (u8)res[0]; \
t0++; \
if (res[1] <= 0) \
   *t1 = 0; \
else if (res[1] >= 255) \
   *t1 = 255; \
else \
   *t1 = (u8)res[1]; \
t1++; \
if (res[2] <= 0) \
   *t2 = 0; \
else if (res[2] >= 255) \
   *t2 = 255; \
else \
   *t2 = (u8)res[2]; \
t2++; \
if (res[3] <= 0) \
   *t3 = 0; \
else if (res[3] >= 255) \
   *t3 = 255; \
else \
   *t3 = (u8)res[3]; \
t3++


/* Does the overlap filtering operation in the horizontal direction. */
void VC1_overlap_X_last(s16 *blk, s16 *pre, u8 *out2, int inc)
{
   u8  *out0, *out1, *out3;
   s16  res[4];
   u16  cols;

   if (out2) {
      out2 -=  8;
# ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("Poverlap_X_last()\n");
      VC1_TRACE_END_BLK
# endif
      out1 = out2 - inc;
      out3 = out2 + inc;
      out0 = out1 - inc;
      cols = 4;
      do {
         /* even cols */
         OVERLAP_OP4(out0, out1, out2, out3, 4, 3);

         /* odd cols */
         OVERLAP_OP4(out0, out1, out2, out3, 3, 4);
      } while (--cols);
      pre -= 32-2;
   }
   else
      pre += 2;

   /* store top edge */
   cols = 8;
   do {
      *(pre++) = blk[0];
      *(pre++) = blk[8];
      blk++;
      pre += 2;
   } while (--cols);
}

#else

/*# define OVERLAP_OFFSET(p) ((p) ? ((p == 255) ? (128-31) : 128) : (128+32))*/
# define OVERLAP_OFFSET(p) 128

# define OVERLAP_OP2(t0, t1, c0, c1, i, ii) \
tmp[0] = (s16)t0 - OVERLAP_OFFSET(t0); \
tmp[1] = (s16)t1 - OVERLAP_OFFSET(t1); \
OVERLAP_TRACE_INPUT(tmp[0], tmp[1], blk[0], blk[i], c0, c1) \
res[0] = c0+1024;  res[1] = c1+1024;  res[2] = c0;       res[3] = c1; \
res[0]+= 7*tmp[0]; res[1]-=   tmp[0]; res[2]+=   tmp[0]; res[3]+=   tmp[0]; \
                   res[1]+= 7*tmp[1]; res[2]+=   tmp[1]; \
                   res[1]+=   blk[0]; res[2]+= 7*blk[0]; \
res[0]+=   blk[i]; res[1]+=   blk[i]; res[2]-=   blk[i]; res[3]+= 7*blk[i]; \
res[0]>>= 3;       res[1]>>= 3;       res[2]>>= 3;       res[3]>>= 3; \
OVERLAP_TRACE_OUTPUT((res[0]-128), (res[1]-128), res[2], res[3]) \
blk[0] = res[2]; \
blk[i] = res[3]; \
blk   += ii; \
if (res[0] <= 0) \
   t0 = 0; \
else if (res[0] >= 255) \
   t0 = 255; \
else \
   t0 = (u8)res[0]; \
if (res[1] <= 0) \
   t1 = 0; \
else if (res[1] >= 255) \
   t1 = 255; \
else \
   t1 = (u8)res[1]


/* Does the overlap filtering operation in the vertical direction. */
void VC1_overlap_Y(s16 *blk, u8 *outp, int inc)
{
   s16 tmp[2];
   s16 res[4];
   u16 rows = 4;

# ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("overlap_Y()\n");
   VC1_TRACE_END_BLK
# endif
   outp -= 2;
   do {
      /* even rows */
      OVERLAP_OP2(outp[0], outp[1], 4, 3, 1, 8);
      outp += inc;

      /* odd rows */
      OVERLAP_OP2(outp[0], outp[1], 3, 4, 1, 8);
      outp += inc;
   } while (--rows);
}

/* Does the overlap filtering operation in the horizontal direction. */
void VC1_overlap_X(s16 *blk, u8 *out0, int inc)
{
   u8  *out1;
   s16  tmp[2];
   s16  res[4];
   u16  cols = 4;

# ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("overlap_X()\n");
   VC1_TRACE_END_BLK
# endif
   out1 = out0 - inc;
   out0 = out1 - inc;
   do {
      /* even cols */
      OVERLAP_OP2(out0[0], out1[0], 4, 3, 8, 1);
      out0++;
      out1++;

      /* odd cols */
      OVERLAP_OP2(out0[0], out1[0], 3, 4, 8, 1);
      out0++;
      out1++;
   } while (--cols);
}


# define OVERLAP_OP4(t0, t1, t2, t3, c0, c1) \
tmp[0] = (s16)(*t0) - OVERLAP_OFFSET(*t0); \
tmp[1] = (s16)(*t1) - OVERLAP_OFFSET(*t1); \
tmp[2] = (s16)(*t2) - OVERLAP_OFFSET(*t2); \
tmp[3] = (s16)(*t3) - OVERLAP_OFFSET(*t3); \
OVERLAP_TRACE_INPUT(tmp[0], tmp[1], tmp[2], tmp[3], c0, c1) \
res[0] = c0+1024;  res[1] = c1+1024;  res[2] = c0+1024;  res[3] = c1+1024; \
res[0]+= 7*tmp[0]; res[1]-=   tmp[0]; res[2]+=   tmp[0]; res[3]+=   tmp[0]; \
                   res[1]+= 7*tmp[1]; res[2]+=   tmp[1]; \
                   res[1]+=   tmp[2]; res[2]+= 7*tmp[2]; \
res[0]+=   tmp[3]; res[1]+=   tmp[3]; res[2]-=   tmp[3]; res[3]+= 7*tmp[3]; \
res[0]>>= 3;       res[1]>>= 3;       res[2]>>= 3;       res[3]>>= 3; \
OVERLAP_TRACE_OUTPUT((res[0]-128), (res[1]-128), (res[2]-128), (res[3]-128)) \
if (res[0] <= 0) \
   *t0 = 0; \
else if (res[0] >= 255) \
   *t0 = 255; \
else \
   *t0 = (u8)res[0]; \
t0++; \
if (res[1] <= 0) \
   *t1 = 0; \
else if (res[1] >= 255) \
   *t1 = 255; \
else \
   *t1 = (u8)res[1]; \
t1++; \
if (res[2] <= 0) \
   *t2 = 0; \
else if (res[2] >= 255) \
   *t2 = 255; \
else \
   *t2 = (u8)res[2]; \
t2++; \
if (res[3] <= 0) \
   *t3 = 0; \
else if (res[3] >= 255) \
   *t3 = 255; \
else \
   *t3 = (u8)res[3]; \
t3++


/* Does the overlap filtering operation in the horizontal direction. */
void VC1_overlap_X_last(u8 *out2, int inc)
{
   u8  *out0, *out1, *out3;
   s16  tmp[4];
   s16  res[4];
   u16  cols = 4;

   out2 -= 8;
# ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("overlap_X_last()\n");
   VC1_TRACE_END_BLK
# endif
   out1 = out2 - inc;
   out3 = out2 + inc;
   out0 = out1 - inc;
   do {
      /* even cols */
      OVERLAP_OP4(out0, out1, out2, out3, 4, 3);

      /* odd cols */
      OVERLAP_OP4(out0, out1, out2, out3, 3, 4);
   } while (--cols);
}
#endif
