/***************************************************************************
                            wmv.c  -  description
                             -------------------
   Stuff needed for handling wmv demuxing.
                             -------------------
    begin                : Thu May 25 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define WMV_TRACE*/
#ifdef WMV_TRACE
# include <stdio.h>
#endif
#include <string.h>

#include "wmv.h"
#include "bitplane.h"
#include "bitstream.h"

#define DATA_PTR        ptr
#define END_DATA_PTR    end_pos
#define PARSER_STATE    wmv->state


/* Read the next GUID.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_next_guid(WMV_DATA *wmv)
{
   const u8 *ptr     = wmv->payload,
            *end_pos = wmv->payload + wmv->payload_len;

#undef  STOP_PARSING
#define STOP_PARSING    goto WMV_next_guid_END

   while (PARSER_STATE < (16+8)) {
#ifdef WMV_TRACE
      if (PARSER_STATE < 16)
         printf("%02hX ", (u16)(*DATA_PTR));
#endif

      switch (PARSER_STATE) {
      case  0:
         switch (*DATA_PTR) {
         case 0x30:
            wmv->guid = WMV_GUID_HEADER;
            break;
         case 0x36:
            wmv->guid = WMV_GUID_DATA;
            break;
         case 0x91:
            wmv->guid = WMV_GUID_STREAM;
            break;
         case 0xA1:
            wmv->guid = WMV_GUID_FILE;
            break;
         default:
            wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case  1:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0x26)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
            if (*DATA_PTR != 0x07)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0xDC)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case  2:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0xB2)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
            if (*DATA_PTR != 0xDC)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0xAB)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case  3:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0x75)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
            if (*DATA_PTR != 0xB7)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0x8C)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case  4:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0x8E)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
            if (*DATA_PTR != 0xB7)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0x47)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case  5:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0x66)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0xA9)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case  6:
         if ( (wmv->guid != WMV_skip_guid_STATE)&&(*DATA_PTR != 0xCF) )
            wmv->guid = WMV_skip_guid_STATE;
         break;
      case  7:
         if ( (wmv->guid != WMV_skip_guid_STATE)&&(*DATA_PTR != 0x11) )
            wmv->guid = WMV_skip_guid_STATE;
         break;
      case  8:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0xA6)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0x8E)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case  9:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0xD9)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
            if (*DATA_PTR != 0xE6)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0xE4)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case 10:
         if ( (wmv->guid != WMV_skip_guid_STATE)&&(*DATA_PTR) )
            wmv->guid = WMV_skip_guid_STATE;
         break;
      case 11:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0xAA)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0xC0)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case 12:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0x0C)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case 13:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0x62)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0x20)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case 14:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0xCE)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0x53)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case 15:
         switch (wmv->guid) {
         case WMV_GUID_HEADER:
         case WMV_GUID_DATA:
            if (*DATA_PTR != 0x6C)
               wmv->guid = WMV_skip_guid_STATE;
            break;
         case WMV_GUID_STREAM:
         case WMV_GUID_FILE:
            if (*DATA_PTR != 0x65)
               wmv->guid = WMV_skip_guid_STATE;
         }
         break;
      case 16:
         wmv->size  = (u64)(*DATA_PTR);
         break;
      default:
         wmv->size |= (u64)(*DATA_PTR) << ((PARSER_STATE-16) << 3);
      }

      PARSER_STATE++;
      NEXT_BYTE(PARSER_STATE)
   }

   if (wmv->guid == WMV_GUID_HEADER) {
      wmv->size      = 30;
      wmv->FH_flags  =  0;
      wmv->stream_id =  0;
   }
   else if ( ( (wmv->guid == WMV_GUID_FILE)  &&(wmv->size < 104) )||
             ( (wmv->guid == WMV_GUID_STREAM)&&(wmv->size < 133) )||
             ( (wmv->guid == WMV_GUID_DATA)&&((!wmv->FH_flags)||
                                              (!wmv->stream_id)) ) )
      wmv->guid = WMV_skip_guid_STATE;

#ifdef WMV_TRACE
   switch (wmv->guid) {
   case WMV_GUID_HEADER:
      printf("WMV_HEADER");
      break;
   case WMV_GUID_DATA:
      printf("WMV_DATA  ");
      break;
   case WMV_GUID_STREAM:
      printf("STREAM_HDR");
      break;
   case WMV_GUID_FILE:
      printf("FILE_HDR  ");
      break;
   default:
      printf("ignored   ");
   }
   printf(" %10ubytes\n", (u32)wmv->size);
#endif

   if (wmv->size > (16+8))
      wmv->size -= 16+8;
   else
      wmv->size  = 0;

   PARSER_STATE      = wmv->guid;
   wmv->payload_len -= (u32)(DATA_PTR - wmv->payload);
   wmv->payload      = DATA_PTR;
   return;

WMV_next_guid_END:
   wmv->payload_len = 0;
}

/* Read the FILE_HEADER.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_read_file(WMV_DATA *wmv)
{
   const u8 *ptr     = wmv->payload,
            *end_pos = wmv->payload + wmv->payload_len;
   u32       pos     = PARSER_STATE - WMV_GUID_FILE;

#undef  STOP_PARSING
#define STOP_PARSING    goto WMV_read_file_END

   while (pos < 76) {
      if (!pos) {
         wmv->FH_pkts     = 0;
         wmv->FH_flags    =
         wmv->FH_min_pkt  =
         wmv->FH_max_pkt  = 0;
      }
      else if ( (32 <= pos)&&(pos < 40) )
         wmv->FH_pkts    |= (u64)(*DATA_PTR) << ((pos-32) << 3);
      else if ( (64 <= pos)&&(pos < 68) )
         wmv->FH_flags   |= (u32)(*DATA_PTR) << ((pos-64) << 3);
      else if ( (68 <= pos)&&(pos < 72) )
         wmv->FH_min_pkt |= (u32)(*DATA_PTR) << ((pos-68) << 3);
      else if   (72 <= pos)
         wmv->FH_max_pkt |= (u32)(*DATA_PTR) << ((pos-72) << 3);
      pos++;
      PARSER_STATE++;
      NEXT_BYTE(PARSER_STATE)
   }

#ifdef WMV_TRACE
   printf("   pkts    = %u\n"
          "   flags   = %u\n"
          "   min_pkt = %ubytes\n"
          "   max_pkt = %ubytes\n",
          (u32)wmv->FH_pkts,
          wmv->FH_flags,
          wmv->FH_min_pkt,
          wmv->FH_max_pkt);
#endif

   wmv->size        -= (u64)(DATA_PTR - wmv->payload);
   wmv->payload_len -= (u32)(DATA_PTR - wmv->payload);
   wmv->payload      = DATA_PTR;
   PARSER_STATE      = (wmv->size) ?
                       WMV_skip_guid_STATE :
                       WMV_next_guid_STATE;
   return;

WMV_read_file_END:
   wmv->size       -= (u64)wmv->payload_len;
   wmv->payload_len = 0;
}

/* Read the STREAM_HEADER.
      INPUT: wmv - Pointer to a WMV_DATA structure.
             seq - Pointer to a SEQ_LAYER structure. */
void WMV_read_stream(WMV_DATA *wmv, SEQ_LAYER *seq)
{
   static const u8 video_guid[16] = {0xC0, 0xEF, 0x19, 0xBC,
                                     0x4D, 0x5B, 0xCF, 0x11,
                                     0xA8, 0xFD, 0x00, 0x80,
                                     0x5F, 0x5C, 0x44, 0x2B};
   const u8 *ptr     = wmv->payload,
            *end_pos = wmv->payload + wmv->payload_len;
   u32       pos     = PARSER_STATE - WMV_GUID_STREAM;

#undef  STOP_PARSING
#define STOP_PARSING    goto WMV_read_stream_END

   while (pos < 105) {
      if ( (pos < 16)&&(*DATA_PTR != video_guid[pos]) )
         break;   /* incorrect guid */
      else if (!pos) {
         wmv->SH_size = 0;
         memset(seq, 0, sizeof(SEQ_LAYER));
      }
      else if (pos == 48)
         wmv->stream_id = *DATA_PTR & 0x7F;
      else if ( (65 <= pos)&&(pos < 69) )
         wmv->SH_size |= (u32)(*DATA_PTR) << ((pos-65) << 3);
      else if ( (69 == pos)||(pos == 70) )
         seq->wid     |= (u16)(*DATA_PTR) << ((pos-69) << 3);
      else if ( (73 == pos)||(pos == 74) )
         seq->hgt     |= (u16)(*DATA_PTR) << ((pos-73) << 3);
      else if ( (71 <= pos)&&(pos < 77)&&(*DATA_PTR) )
         break;   /* frame_hgt or frame_wid values are too big */
      else if ( ((pos == 81)&&(*DATA_PTR != 'W'))||
                ((pos == 82)&&(*DATA_PTR != 'M'))||
                ((pos == 83)&&(*DATA_PTR != 'V'))||
                ((pos == 84)&&(*DATA_PTR != '3') /*&&(*DATA_PTR != '2')*/ ) )
         break;   /* incorrect fourcc */
      pos++;
      PARSER_STATE++;
      NEXT_BYTE(PARSER_STATE)
   }

   if ( (seq->wid)&&(seq->wid & 0x000F) ) {
      seq->wid &= 0xFFF0;
      seq->wid += 0x0010;
   }
   if ( (seq->hgt)&&(seq->hgt & 0x000F) ) {
      seq->hgt &= 0xFFF0;
      seq->hgt += 0x0010;
   }

   if ( (pos == 105)&&(wmv->SH_size >= 44)&&
        (seq->wid)&&(seq->wid <= VC1_MAX_FRAME_WID)&&
        (seq->hgt)&&(seq->hgt <= VC1_MAX_FRAME_HGT) ) {
      if (!seq->disp_wid) {
         seq->disp_wid = seq->wid;
         seq->disp_hgt = seq->hgt;
#ifdef WMV_TRACE
         printf("   stream_id  = %hu\n"
                "   hdr_size   = %ubytes\n"
                "   frame_size = %hux%hu\n",
               (u16)wmv->stream_id,
               wmv->SH_size,
               seq->wid, seq->hgt);
#endif
      }
      seq->payload     = DATA_PTR;
      seq->payload_len = wmv->payload_len -
                         (u32)(DATA_PTR - wmv->payload);
      VC1_VIDEO_read_SEQ_STRUCT(seq);
      if (seq->state < VC1_VIDEO_read_ENTRYPT_LAYER_STATE)
         goto WMV_read_stream_END;
      DATA_PTR = seq->payload;

      if ( (VC1_SEQ_PROFILEp(seq) > VC1_PROFILE_MAIN)||
           (VC1_SEQ_LEVELp(seq))||
           (VC1_SEQ_MULTIRESp(seq))||
           (VC1_SEQ_SYNC_MARKERSp(seq)) )
         wmv->stream_id = 0;  /* do not try decoding this stream */
      wmv->frame_size  =
      wmv->frag_offset = 0;
   }
   else
      wmv->stream_id = 0;

   wmv->size        -= (u64)(DATA_PTR - wmv->payload);
   wmv->payload_len -= (u32)(DATA_PTR - wmv->payload);
   wmv->payload      = DATA_PTR;
   PARSER_STATE      = (wmv->size) ?
                       WMV_skip_guid_STATE :
                       WMV_next_guid_STATE;
   return;

WMV_read_stream_END:
   wmv->size       -= (u64)wmv->payload_len;
   wmv->payload_len = 0;
}

/* Read the WMV_DATA.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_read_data(WMV_DATA *wmv)
{
   const u8 *ptr     = wmv->payload,
            *end_pos = wmv->payload + wmv->payload_len;

#undef  STOP_PARSING
#define STOP_PARSING    goto WMV_read_data_END

#define START_PKT       (WMV_GUID_DATA+26)
#define SBYTES          (WMV_GUID_DATA+27)
#define PKT_SIZE        (WMV_GUID_DATA+28)
#define SEQ_SIZE        (WMV_GUID_DATA+29)
#define PAD_SIZE        (WMV_GUID_DATA+30)
#define TIMESTAMP       (WMV_GUID_DATA+31)
#define START_FRAGS     (WMV_GUID_DATA+32)
#define STREAM_ID       (WMV_GUID_DATA+33)
#define FRAG_SBYTES     (WMV_GUID_DATA+34)
#define FRAG_SEQ        (WMV_GUID_DATA+35)
#define FRAG_OFFSET     (WMV_GUID_DATA+36)
#define FRAG_EXT_SIZE   (WMV_GUID_DATA+37)
#define FRAG_EXT        (WMV_GUID_DATA+38)
#define FRAG_SIZE       (WMV_GUID_DATA+39)
#define END_FRAG        (WMV_GUID_DATA+41)
#define END_PKT         (WMV_GUID_DATA+42)
   /* skip the first 26 bytes */
   while (PARSER_STATE < START_PKT) {
      PARSER_STATE++;
      NEXT_BYTE(PARSER_STATE)
   }
   
   switch (PARSER_STATE) {
   case START_PKT:
      goto WMV_read_data_START_PKT;
   case SBYTES:
      goto WMV_read_data_SBYTES;
   case PKT_SIZE:
      goto WMV_read_data_PKT_SIZE;
   case SEQ_SIZE:
      goto WMV_read_data_SEQ_SIZE;
   case PAD_SIZE:
      goto WMV_read_data_PAD_SIZE;
   case TIMESTAMP:
      goto WMV_read_data_TIMESTAMP;
   case START_FRAGS:
      goto WMV_read_data_START_FRAGS;
   case STREAM_ID:
      goto WMV_read_data_STREAM_ID;
   case FRAG_SBYTES:
      goto WMV_read_data_FRAG_SBYTES;
   case FRAG_SEQ:
      goto WMV_read_data_FRAG_SEQ;
   case FRAG_OFFSET:
      goto WMV_read_data_FRAG_OFFSET;
   case FRAG_EXT_SIZE:
      goto WMV_read_data_FRAG_EXT_SIZE;
   case FRAG_EXT:
      goto WMV_read_data_FRAG_EXT;
   case FRAG_SIZE:
      goto WMV_read_data_FRAG_SIZE;
   case WMV_END_FRAME_STATE:
   case END_FRAG:
      goto WMV_read_data_END_FRAG;
   case END_PKT:
      goto WMV_read_data_END_PKT;
   }

WMV_read_data_START_PKT:
   if (*DATA_PTR != 0x82) {
#ifdef WMV_TRACE
      printf("First byte in packet (0x%02hX) is not equal to 0x82\n",
             (u16)(*DATA_PTR));
#endif
      goto WMV_read_data_NO_MORE_DATA;
   }
   wmv->SH_size     =
   wmv->WD_sbytes   =  0;
   wmv->WD_hdr_size = 11;
   NEXT_BYTE(SBYTES)

WMV_read_data_SBYTES:
   while (wmv->SH_size < 32) {
      wmv->WD_sbytes |= (u32)(*DATA_PTR) << wmv->SH_size;
      wmv->SH_size   += 8;
      NEXT_BYTE(SBYTES)
   }
   if (wmv->WD_sbytes & 0x0000FFFF) {
#ifdef WMV_TRACE
      printf("Second or third bytes in packet (0x%04X) is not equal to 0\n",
             (wmv->WD_sbytes & 0x0000FFFF));
#endif
      goto WMV_read_data_NO_MORE_DATA;
   }

   wmv->WD_sbytes |= (wmv->WD_sbytes >> 18) & 0x00000018;
   if (wmv->WD_sbytes & 0x00000FFF) {
      if ((wmv->WD_sbytes & 0x00000FFF) == 0x00000018)
         wmv->WD_sbytes += 8;
      wmv->SH_size     =
      wmv->WD_pkt_size = 0;

WMV_read_data_PKT_SIZE:
      while (wmv->SH_size < (wmv->WD_sbytes & 0x00000FFF)) {
         wmv->WD_pkt_size |= (u32)(*DATA_PTR) << wmv->SH_size;
         wmv->SH_size     += 8;
         wmv->WD_hdr_size++;
         NEXT_BYTE(PKT_SIZE)
      }
      wmv->WD_sbytes &= 0xFFFFF000;
   }
   else
      wmv->WD_pkt_size = wmv->FH_max_pkt;

   /* ignore seq_size */   
   wmv->WD_sbytes |= (wmv->WD_sbytes >> 14) & 0x00000018;
   if (wmv->WD_sbytes & 0x00000FFF) {
      if ((wmv->WD_sbytes & 0x00000FFF) == 0x00000018)
         wmv->WD_sbytes += 8;
      wmv->SH_size = 0;

WMV_read_data_SEQ_SIZE:
      while (wmv->SH_size < (wmv->WD_sbytes & 0x00000FFF)) {
         wmv->WD_pad_size |= (u32)(*DATA_PTR) << wmv->SH_size;
         wmv->SH_size     += 8;
         wmv->WD_hdr_size++;
         NEXT_BYTE(SEQ_SIZE)
      }
      wmv->WD_sbytes &= 0xFFFFF000;
   }

   wmv->WD_pad_size = 0;
   wmv->WD_sbytes  |= (wmv->WD_sbytes >> 16) & 0x00000018;
   if (wmv->WD_sbytes & 0x00000FFF) {
      if ((wmv->WD_sbytes & 0x00000FFF) == 0x00000018)
         wmv->WD_sbytes += 8;
      wmv->SH_size = 0;

WMV_read_data_PAD_SIZE:
      while (wmv->SH_size < (wmv->WD_sbytes & 0x00000FFF)) {
         wmv->WD_pad_size |= (u32)(*DATA_PTR) << wmv->SH_size;
         wmv->SH_size     += 8;
         wmv->WD_hdr_size++;
         NEXT_BYTE(PAD_SIZE)
      }
      wmv->WD_sbytes &= 0xFFFFF000;
   }

   /* ignore timestamp, and duration values */
   wmv->SH_size = 6;
WMV_read_data_TIMESTAMP:
   while (wmv->SH_size) {
      wmv->SH_size--;
      NEXT_BYTE(TIMESTAMP)
   }

   if (wmv->WD_sbytes & 0x00010000) {
      wmv->WD_sbytes |= (u32)(*DATA_PTR & 0xC0) << 8;
      wmv->WD_frags   = (u32)(*DATA_PTR & 0x3F);
      wmv->WD_hdr_size++;
      NEXT_BYTE(START_FRAGS)
   }
   else {
      wmv->WD_sbytes |= 0x00008000;
      wmv->WD_frags   = 1;
   }

WMV_read_data_START_FRAGS:
   wmv->WD_hdr_size += wmv->WD_pad_size;
   if (wmv->FH_min_pkt > wmv->WD_pkt_size)
      wmv->WD_pad_size += wmv->FH_min_pkt - wmv->WD_pkt_size;
   wmv->WD_pkt_size -= wmv->WD_hdr_size;
   
#ifdef WMV_TRACE
   printf("<WMV_DATA_PACKET[%u]>\n", (wmv->FH_pkts - 1));
   printf("   sbytes   = 0x%08X\n"
          "   pkt_size = %ubytes\n"
          "   pad_size = %ubytes\n"
          "   frags    = %u\n",
          wmv->WD_sbytes,
          wmv->WD_pkt_size,
          wmv->WD_pad_size,
          wmv->WD_frags);
#endif

WMV_read_data_STREAM_ID:
   wmv->WD_hdr_size = 1;
   wmv->cur_id      = *DATA_PTR & 0x7F;
   NEXT_BYTE(FRAG_SBYTES)

WMV_read_data_FRAG_SBYTES:
   /* ignore frag_seq */
   wmv->WD_sbytes |= (wmv->WD_sbytes >> 25) & 0x00000018;
   if (wmv->WD_sbytes & 0x00000FFF) {
      if ((wmv->WD_sbytes & 0x00000FFF) == 0x00000018)
         wmv->WD_sbytes += 8;
      wmv->SH_size = 0;

WMV_read_data_FRAG_SEQ:
      while (wmv->SH_size < (wmv->WD_sbytes & 0x00000FFF)) {
         wmv->WD_frag_size |= (u32)(*DATA_PTR) << wmv->SH_size;
         wmv->SH_size      += 8;
         wmv->WD_hdr_size++;
         NEXT_BYTE(FRAG_SEQ)
      }
      wmv->WD_sbytes &= 0xFFFFF000;
   }

   wmv->WD_frag_offset = 0;
   wmv->WD_sbytes     |= (wmv->WD_sbytes >> 23) & 0x00000018;
   if (wmv->WD_sbytes & 0x00000FFF) {
      if ((wmv->WD_sbytes & 0x00000FFF) == 0x00000018)
         wmv->WD_sbytes += 8;
      wmv->SH_size = 0;

WMV_read_data_FRAG_OFFSET:
      while (wmv->SH_size < (wmv->WD_sbytes & 0x00000FFF)) {
         wmv->WD_frag_offset |= (u32)(*DATA_PTR) << wmv->SH_size;
         wmv->SH_size        += 8;
         wmv->WD_hdr_size++;
         NEXT_BYTE(FRAG_OFFSET)
      }
      wmv->WD_sbytes &= 0xFFFFF000;
   }

   wmv->WD_frag_size = 0;
   wmv->WD_sbytes   |= (wmv->WD_sbytes >> 21) & 0x00000018;
   if (wmv->WD_sbytes & 0x00000FFF) {
      if ((wmv->WD_sbytes & 0x00000FFF) == 0x00000018)
         wmv->WD_sbytes += 8;
      wmv->SH_size = 0;

WMV_read_data_FRAG_EXT_SIZE:
      while (wmv->SH_size < (wmv->WD_sbytes & 0x00000FFF)) {
         wmv->WD_frag_size |= (u32)(*DATA_PTR) << wmv->SH_size;
         wmv->SH_size      += 8;
         wmv->WD_hdr_size++;
         NEXT_BYTE(FRAG_EXT_SIZE)
      }
      wmv->WD_sbytes &= 0xFFFFF000;
   }
   if (wmv->WD_frag_size < 8) {
#ifdef WMV_TRACE
      printf("Fragment header extension is too small (%u)\n",
             wmv->WD_frag_size);
#endif
      wmv->WD_pkt_size -= wmv->WD_hdr_size;
      goto WMV_read_data_END_PKT;
   }
   wmv->WD_hdr_size  += wmv->WD_frag_size;
   wmv->WD_frame_size = 
   wmv->SH_size       = 0;

WMV_read_data_FRAG_EXT:
   while (wmv->WD_frag_size) {
      wmv->WD_frag_size--;
      if (wmv->SH_size < 32)
         wmv->WD_frame_size |= (u32)(*DATA_PTR) << wmv->SH_size;
      wmv->SH_size += 8;
      NEXT_BYTE(FRAG_EXT)
   }

   wmv->WD_pkt_size -= wmv->WD_hdr_size;
   if (wmv->WD_sbytes & 0x00010000) {
      wmv->WD_sbytes |= (wmv->WD_sbytes >> 11) & 0x00000018;
      if (wmv->WD_sbytes & 0x00000FFF) {
         if ((wmv->WD_sbytes & 0x00000FFF) == 0x00000018)
            wmv->WD_sbytes += 8;
         wmv->SH_size = 0;

WMV_read_data_FRAG_SIZE:
         while (wmv->SH_size < (wmv->WD_sbytes & 0x00000FFF)) {
            wmv->WD_frag_size |= (u32)(*DATA_PTR) << wmv->SH_size;
            wmv->SH_size      += 8;
            wmv->WD_pkt_size--;
            NEXT_BYTE(FRAG_SIZE)
         }
         wmv->WD_sbytes &= 0xFFFFF000;
      }
      wmv->WD_pkt_size -= wmv->WD_frag_size;
   }
   else if (wmv->WD_frags != 1) {
#ifdef WMV_TRACE
      printf("Missing fragment size\n");
#endif
      goto WMV_read_data_END_PKT;
   }
   else {
      wmv->WD_frag_size = wmv->WD_pkt_size;
      wmv->WD_pkt_size  = 0;
   }

   if (wmv->cur_id == wmv->stream_id) {
      if (!wmv->WD_frag_offset) {
         if (wmv->skip_frames) {
            wmv->skip_frames--;
            goto WMV_read_data_END_FRAG;
         }
#ifdef WMV_TRACE         
         else if (wmv->frame_size > wmv->frag_offset) {
            printf("Missing final %u frame bytes\n", 
                   (wmv->frame_size - wmv->frag_offset));
         }
#endif
         /* start new frame */
         wmv->frame_size  = wmv->WD_frame_size;
         wmv->frag_offset = 0;
      }
      else if (wmv->skip_frames)
         goto WMV_read_data_END_FRAG;

      wmv->size        -= (u64)(DATA_PTR - wmv->payload);
      wmv->payload_len -= (u32)(DATA_PTR - wmv->payload);
      wmv->payload      = DATA_PTR;
#ifdef WMV_TRACE
      printf("   <WMV_DATA_FRAGMENT[%u]>\n", (wmv->WD_frags - 1));
      printf("      stream_id  = 0x%02hX\n"
             "      frame_size = %ubytes\n"
             "      fragment   = %u -> %u\n",
             (u16)wmv->cur_id,
             wmv->WD_frame_size,
             wmv->WD_frag_offset, 
             (wmv->WD_frag_offset + wmv->WD_frag_size));
#endif
      
      if ( (wmv->WD_frame_size  == wmv->frame_size)&&
           (wmv->WD_frag_offset == wmv->frag_offset) ) {
         wmv->frag_offset += wmv->WD_frag_size;
         if (wmv->frame_size >= wmv->frag_offset) {
            PARSER_STATE = WMV_FRAME_FRAG_STATE;
            wmv->size   -= (u64)wmv->WD_frag_size;
            return;  /* let decoder deal with the fragment */
         }
#ifdef WMV_TRACE
         else {
            printf("Frame has %u too many bytes\n", 
                   (wmv->frag_offset - wmv->frame_size));
         }
#endif
      }
#ifdef WMV_TRACE
      else if (wmv->WD_frag_offset > wmv->frag_offset) {
         printf("Missing %u frame bytes\n", 
                (wmv->WD_frag_offset - wmv->frag_offset));
      }
      else
         printf("Corrupt frame\n");
#endif
      wmv->frame_size  =
      wmv->frag_offset = 0;
      PARSER_STATE     = WMV_END_FRAME_STATE;
      return;  /* tell decoder to abandon the current frame */
   }

WMV_read_data_END_FRAG:
   if (wmv->WD_frag_size) {
      wmv->size        -= (u64)(DATA_PTR - wmv->payload);
      wmv->payload_len -= (u32)(DATA_PTR - wmv->payload);
      wmv->payload      = DATA_PTR;
      if (wmv->payload_len <= wmv->WD_frag_size) {
         wmv->WD_frag_size -= wmv->payload_len;
         PARSER_STATE       = END_FRAG;
         goto WMV_read_data_END;
      }
      DATA_PTR += wmv->WD_frag_size;
   }
   if (--(wmv->WD_frags))
      goto WMV_read_data_STREAM_ID;

WMV_read_data_END_PKT:
   if (wmv->WD_pad_size) {
      wmv->WD_pkt_size += wmv->WD_pad_size;
      wmv->WD_pad_size  = 0;
   }
   if (wmv->WD_pkt_size) {
      wmv->size        -= (u64)(DATA_PTR - wmv->payload);
      wmv->payload_len -= (u32)(DATA_PTR - wmv->payload);
      wmv->payload      = DATA_PTR;
      if (wmv->payload_len <= wmv->WD_pkt_size) {
         wmv->WD_pkt_size -= wmv->payload_len;
         PARSER_STATE      = END_PKT;
         goto WMV_read_data_END;
      }
      DATA_PTR += wmv->WD_pkt_size;
   }
   if (--(wmv->FH_pkts))
      goto WMV_read_data_START_PKT;

WMV_read_data_NO_MORE_DATA:
#ifdef WMV_TRACE
   printf("<WMV_DATA_END>\n");
#endif
   wmv->size        -= (u64)(DATA_PTR - wmv->payload);
   wmv->payload_len -= (u32)(DATA_PTR - wmv->payload);
   wmv->payload      = DATA_PTR;
   PARSER_STATE      = (wmv->size) ? 
                       WMV_skip_guid_STATE :
                       WMV_next_guid_STATE;
   return;

WMV_read_data_END:
   wmv->size       -= (u64)wmv->payload_len;
   wmv->payload_len = 0;

#undef START_PKT
#undef SBYTES
#undef PKT_SIZE
#undef SEQ_SIZE
#undef PAD_SIZE
#undef TIMESTAMP
#undef START_FRAGS
#undef STREAM_ID
#undef FRAG_SBYTES
#undef FRAG_SEQ
#undef FRAG_OFFSET
#undef FRAG_EXT_SIZE
#undef FRAG_EXT
#undef FRAG_SIZE
#undef END_FRAG
#undef END_PKT
}

/* Skip the current GUID.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_skip_guid(WMV_DATA *wmv)
{
   if ((u64)wmv->payload_len < wmv->size) {
      wmv->size       -= (u64)wmv->payload_len;
      wmv->payload_len = 0;
      return;
   }
   wmv->payload     += (u32)wmv->size;
   wmv->payload_len -= (u32)wmv->size;
   PARSER_STATE      = WMV_next_guid_STATE;
}

